package ru.dbotthepony.mc.otm.block.entity

import net.minecraft.core.BlockPos
import net.minecraft.core.Direction
import net.minecraft.nbt.ByteArrayTag
import net.minecraft.nbt.CompoundTag
import net.minecraft.nbt.IntTag
import net.minecraft.nbt.ListTag
import net.minecraft.network.chat.Component
import net.minecraft.network.chat.TranslatableComponent
import net.minecraft.server.level.ServerLevel
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.AbstractContainerMenu
import net.minecraft.world.level.Level
import net.minecraft.world.level.block.state.BlockState
import net.minecraftforge.common.capabilities.Capability
import net.minecraftforge.common.util.LazyOptional
import net.minecraftforge.energy.CapabilityEnergy
import net.minecraftforge.energy.IEnergyStorage
import ru.dbotthepony.mc.otm.*
import ru.dbotthepony.mc.otm.block.BlockEnergyCounter
import ru.dbotthepony.mc.otm.capability.IMatteryEnergyStorage
import ru.dbotthepony.mc.otm.capability.MatteryCapability
import ru.dbotthepony.mc.otm.capability.extractEnergy
import ru.dbotthepony.mc.otm.capability.receiveEnergy
import ru.dbotthepony.mc.otm.core.Fraction
import ru.dbotthepony.mc.otm.menu.EnergyCounterMenu
import java.lang.ref.WeakReference

class BlockEntityEnergyCounter(p_155229_: BlockPos, p_155230_: BlockState) : BlockEntityMattery(Registry.BlockEntities.ENERGY_COUNTER, p_155229_, p_155230_) {
	var passed = Fraction.ZERO
		private set

	private val history = Array(10 * 20) { Fraction.ZERO }
	private var historyTick = 0

	fun size() = history.size
	operator fun get(i: Int) = history[i]
	val lastTick: Fraction get() = history[historyTick]

	fun getHistory(ticks: Int): Array<Fraction> {
		require(!(ticks < 1 || ticks >= history.size)) { "Invalid history length provided" }

		val history = Array(ticks) { Fraction.ZERO }

		for (i in 0 until ticks) {
			var index = (historyTick - i) % this.history.size
			if (index < 0) index += this.history.size
			history[i] = this.history[index]
		}

		return history
	}

	fun calcAverage(ticks: Int): Fraction {
		return sumHistory(ticks) / ticks
	}

	fun sumHistory(ticks: Int): Fraction {
		require(!(ticks < 1 || ticks >= history.size)) { "Invalid history length provided" }

		var value = Fraction.ZERO

		for (i in 0 until ticks) {
			var index = (historyTick - i) % history.size
			if (index < 0) index += history.size
			value += history[index]
		}

		return value
	}

	override fun saveAdditional(nbt: CompoundTag) {
		super.saveAdditional(nbt)
		nbt["passed"] = passed.serializeNBT()

		val list = ListTag()
		nbt["history"] = list
		nbt["history_tick"] = historyTick

		for (num in history)
			list.add(num.serializeNBT())
	}

	override fun load(nbt: CompoundTag) {
		super.load(nbt)

		nbt.ifHas(("passed")) {
			passed = Fraction.deserializeNBT(it)
		}

		nbt.ifHas(("history_tick"), IntTag::class.java) {
			historyTick = it.asInt
		}

		nbt.ifHas("history", ListTag::class.java) {
			for (i in it.indices) {
				val bytes = it[i] as? ByteArrayTag
				history[i] = if (bytes != null) Fraction.deserializeNBT(bytes) else Fraction.ZERO
			}
		}
	}

	override fun getDefaultDisplayName(): Component {
		return NAME
	}

	override fun createMenu(containerID: Int, inventory: Inventory, ply: Player): AbstractContainerMenu {
		return EnergyCounterMenu(containerID, inventory, this)
	}

	private val energyInput = EnergyCounterCap(true)
	private val energyOutput = EnergyCounterCap(false)

	private var inputCapability = LazyOptional.empty<IEnergyStorage>()
	private var outputCapability = LazyOptional.empty<IEnergyStorage>()

	override fun setLevel(p_155231_: Level) {
		super.setLevel(p_155231_)

		val level = level
		if (level is ServerLevel) {
			OverdriveThatMatters.tickOnce(level) { checkSurroundings(level) }
		}
	}

	private inner class EnergyCounterCap(val is_input: Boolean) : IMatteryEnergyStorage {
		override fun extractEnergyOuter(howMuch: Fraction, simulate: Boolean): Fraction {
			return extractEnergyInner(howMuch, simulate)
		}

		override fun extractEnergyInner(howMuch: Fraction, simulate: Boolean): Fraction {
			if (is_input)
				return Fraction.ZERO

			if (inputCapability.isPresent) {
				val it = inputCapability.resolve().get()

				val diff: Fraction

				if (it is IMatteryEnergyStorage) {
					diff = it.extractEnergyOuter(howMuch, simulate)
				} else {
					diff = Fraction(it.extractEnergy(howMuch, simulate))
				}

				if (!simulate) {
					passed += diff
					history[historyTick] += diff
				}

				return diff
			}

			return Fraction.ZERO
		}

		override fun receiveEnergyOuter(howMuch: Fraction, simulate: Boolean): Fraction {
			return receiveEnergyInner(howMuch, simulate)
		}

		override fun receiveEnergyInner(howMuch: Fraction, simulate: Boolean): Fraction {
			if (!is_input)
				return Fraction.ZERO

			if (outputCapability.isPresent) {
				val it = outputCapability.resolve().get()

				val diff: Fraction

				if (it is IMatteryEnergyStorage) {
					diff = it.receiveEnergyOuter(howMuch, simulate)
				} else {
					diff = Fraction(it.receiveEnergy(howMuch, simulate))
				}

				if (!simulate) {
					passed += diff
					history[historyTick] += diff
				}

				return diff
			}

			return Fraction.ZERO
		}

		override val batteryLevel: Fraction
			get() {
				if (is_input) {
					if (outputCapability.isPresent) {
						val it = outputCapability.resolve().get()

						if (it is IMatteryEnergyStorage) {
							return it.batteryLevel
						}

						return Fraction(it.energyStored)
					}
				} else {
					if (inputCapability.isPresent) {
						val it = inputCapability.resolve().get()

						if (it is IMatteryEnergyStorage) {
							return it.batteryLevel
						}

						return Fraction(it.energyStored)
					}
				}

				return Fraction.ZERO
			}

		override val maxBatteryLevel: Fraction
			get() {
				if (is_input) {
					if (outputCapability.isPresent) {
						val it = outputCapability.resolve().get()

						if (it is IMatteryEnergyStorage) {
							return it.maxBatteryLevel
						}

						return Fraction(it.maxEnergyStored)
					}
				} else {
					if (inputCapability.isPresent) {
						val it = inputCapability.resolve().get()

						if (it is IMatteryEnergyStorage) {
							return it.maxBatteryLevel
						}

						return Fraction(it.maxEnergyStored)
					}
				}

				return Fraction.ZERO
			}

		override fun canExtract() = !is_input
		override fun canReceive() = is_input
	}

	private var resolverInput = LazyOptional.of<IMatteryEnergyStorage> { energyInput }
	private var resolverOutput = LazyOptional.of<IMatteryEnergyStorage> { energyOutput }
	private var valid = true

	override fun invalidateCaps() {
		super.invalidateCaps()
		valid = false
		resolverInput.invalidate()
		resolverOutput.invalidate()
	}

	override fun reviveCaps() {
		super.reviveCaps()
		valid = true
		resolverInput = LazyOptional.of { energyInput }
		resolverOutput = LazyOptional.of { energyOutput }
	}

	@Suppress("deprecation")
	override fun setBlockState(new: BlockState) {
		val old = blockState
		super.setBlockState(new)

		if (new !== old && new.getValue(BlockEnergyCounter.INPUT_DIRECTION) != old.getValue(BlockEnergyCounter.INPUT_DIRECTION)) {
			resolverInput.invalidate()
			resolverOutput.invalidate()
			resolverInput = LazyOptional.of { energyInput }
			resolverOutput = LazyOptional.of { energyOutput }

			if (level != null)
				checkSurroundings(level)
		}
	}

	private fun getAndBind(
		level: Level,
		old: LazyOptional<IEnergyStorage>,
		side: Direction
	): LazyOptional<IEnergyStorage> {
		val ent = level.getBlockEntity(blockPos.offset(side.normal)) ?: return LazyOptional.empty()
		val resolve = ent.getCapability(CapabilityEnergy.ENERGY, side.opposite)

		if (resolve !== old) {
			val weak = WeakReference(this)

			resolve.addListener {
				val get = weak.get()

				if (get?.level != null) {
					get.checkSurroundings(get.level)
				}
			}

			return resolve
		}

		return old
	}

	fun checkSurroundings(level: Level?) {
		if (isRemoved || level !is ServerLevel) return

		inputCapability = getAndBind(
			level,
			inputCapability,
			blockState.getValue(BlockEnergyCounter.INPUT_DIRECTION)
		)

		outputCapability = getAndBind(
			level,
			outputCapability,
			-blockState.getValue(BlockEnergyCounter.INPUT_DIRECTION)
		)
	}

	override fun <T> getCapability(cap: Capability<T>, side: Direction?): LazyOptional<T> {
		if (side == null || isRemoved)
			return super.getCapability(cap, side)

		if (valid) {
			if (side == blockState.getValue(BlockEnergyCounter.INPUT_DIRECTION)) {
				if (cap === MatteryCapability.ENERGY || cap === CapabilityEnergy.ENERGY) {
					return resolverInput.cast()
				}
			} else if (side == blockState.getValue(BlockEnergyCounter.INPUT_DIRECTION).opposite) {
				if (cap === MatteryCapability.ENERGY || cap === CapabilityEnergy.ENERGY) {
					return resolverOutput.cast()
				}
			}
		}

		return super.getCapability(cap, side)
	}

	fun tick() {
		historyTick = (historyTick + 1) % history.size
		history[historyTick] = Fraction.ZERO
	}

	companion object {
		private val NAME = TranslatableComponent("block.overdrive_that_matters.energy_counter")
	}
}