package ru.dbotthepony.mc.otm.storage

import ru.dbotthepony.mc.otm.capability.IMatteryEnergyStorage
import ru.dbotthepony.mc.otm.core.Fraction
import java.util.*
import java.util.function.Supplier

class RemoteTuple<T : IStorageStack>(val obj: T, val remote_id: UUID, val provider: IStorageView<T>, val local: LocalTuple<T>) {
	fun extract(amount: Fraction, simulate: Boolean): T {
		return provider.extractStack(remote_id, amount, simulate)
	}

	fun extractCount(amount: Fraction, simulate: Boolean): Fraction {
		return provider.extractStackCount(remote_id, amount, simulate)
	}

	override fun equals(other: Any?): Boolean {
		return other is RemoteTuple<*> && other.remote_id == remote_id || other is UUID && other == remote_id
	}

	override fun hashCode(): Int {
		return remote_id.hashCode()
	}
}

class LocalTuple<T : IStorageStack>(override val stack: T, override val id: UUID, val tuples: ArrayList<RemoteTuple<T>>) : IStorageTuple<T> {
	override fun id(): UUID {
		return id
	}

	override fun stack(): T {
		return stack
	}
}

open class VirtualComponent<T : IStorageStack>(identity: Class<T>) : IStorageComponent<T>, IStorageListener<T> {
	@JvmField
	protected val identity: StorageObjectTuple<T>
	final override fun storageIdentity() = identity.identity()

	init {
		this.identity = StorageObjectRegistry.getOrError(identity)
	}

	// удаленный UUID -> Кортеж
	@JvmField
	protected val remoteByUUID = HashMap<UUID, RemoteTuple<T>>()

	// локальный UUID -> Локальный кортеж
	@JvmField
	protected val localByUUID = HashMap<UUID, LocalTuple<T>>()

	// Хеш ключ -> Список Локальных кортежей
	@JvmField
	protected val partitions = HashMap<Any, ArrayList<LocalTuple<T>>>()

	// ArrayList для скорости работы
	@JvmField
	protected val listeners = ArrayList<IStorageListener<T>>()

	@JvmField
	protected val consumers = ArrayList<IStorageConsumer<T>>()

	open fun add(identity: IStorageIdentity<T>) {
		if (identity is IStorageView<T>) {
			identity.addListenerAuto(this)
		} else if (identity is IStorageTrigger<T>) {
			identity.addListener(this)
		}

		if (identity is IStorageConsumer<T> && !consumers.contains(identity)) {
			consumers.add(identity);
		}
	}

	open fun remove(identity: IStorageIdentity<T>) {
		if (identity is IStorageView<T>) {
			identity.removeListenerAuto(this)
		} else if (identity is IStorageTrigger<T>) {
			identity.removeListener(this)
		}

		if (identity is IStorageConsumer<T>) {
			consumers.remove(identity)
		}
	}

	override fun addListener(listener: IStorageListener<T>): Boolean {
		if (!listeners.contains(listener)) {
			listeners.add(listener)
			return true
		}

		return false
	}

	override fun removeListener(listener: IStorageListener<T>): Boolean {
		return listeners.remove(listener)
	}

	override fun getStack(id: UUID): T {
		return localByUUID[id]?.stack ?: identity.empty
	}

	override fun getStacks(): List<IStorageTuple<T>> {
		var capacity = 0
		for (list in partitions.values) capacity += list.size
		val output = ArrayList<IStorageTuple<T>>(capacity)
		for (listing in partitions.values) output.addAll(listing)
		return output
	}

	override fun addObject(stack: T, id: UUID, provider: IStorageView<T>) {
		check(!remoteByUUID.containsKey(id)) { "Already tracking tuple with id $id" }
		val items = partitions.computeIfAbsent(stack.partitionKey()) { ArrayList() }
		var local: LocalTuple<T>? = null

		for (item in items) {
			if (item.stack.sameItem(stack)) {
				item.stack.grow(stack.count)
				local = item
				break
			}
		}

		val added = local == null

		if (local == null) {
			local = LocalTuple(stack.copy() as T, UUID.randomUUID(), ArrayList<RemoteTuple<T>>(1))
			items.add(local)
			localByUUID[local.id] = local
		}

		val remote = RemoteTuple(stack.copy() as T, id, provider, local)
		local.tuples.add(remote)
		remoteByUUID[id] = remote

		if (added) {
			for (listener in listeners) {
				listener.addObject(local.stack, local.id, this)
			}
		} else {
			for (listener in listeners) {
				listener.changeObject(local.id, local.stack.count)
			}
		}
	}

	override fun changeObject(id: UUID, newCount: Fraction) {
		assert(newCount > Fraction.ZERO)
		val tuple = remoteByUUID[id] ?: throw IllegalStateException("No such tuple with id $id")

		val diff = newCount - tuple.obj.count
		tuple.obj.count = newCount
		tuple.local.stack.grow(diff)

		for (listener in listeners) {
			listener.changeObject(tuple.local.id, tuple.local.stack.count)
		}
	}

	override fun removeObject(id: UUID) {
		val tuple = remoteByUUID[id] ?: throw IllegalStateException("No such tuple with id $id")
		val key = tuple.local.stack.partitionKey()

		tuple.local.stack.shrink(tuple.obj.count)
		tuple.local.tuples.remove(tuple)

		remoteByUUID.remove(id)

		val a = tuple.local.stack.count <= Fraction.ZERO
		val b = tuple.local.tuples.size == 0

		if (a || b) {
			check(a && b) { "View object is empty, but tuple list is not!" }
			localByUUID.remove(tuple.local.id)
			partitions[key]!!.remove(tuple.local)

			for (listener in listeners) {
				listener.removeObject(tuple.local.id)
			}
		}
	}

	override fun insertStack(stack: T, simulate: Boolean): T {
		var leftover = stack

		for (consumer in consumers) {
			leftover = consumer.insertStack(leftover, simulate)

			if (leftover.isEmpty()) {
				return leftover
			}
		}

		return leftover
	}

	override fun extractStack(id: UUID, amount: Fraction, simulate: Boolean): T {
		if (amount.isZero())
			return identity.empty

		var amount = amount
		val tuple: LocalTuple<T>? = localByUUID[id]

		if (tuple == null || amount.isZero())
			return identity.empty

		if (amount <= Fraction.MINUS_ONE)
			amount = tuple.stack.getMaxStackSize().orElse(tuple.stack.count)

		val toExtract = tuple.stack.count.min(amount)
		var extracted = Fraction.ZERO
		val copy = tuple.stack.copy() as T

		for (remote_tuple in tuple.tuples) {
			extracted += remote_tuple.extractCount(toExtract - extracted, simulate)

			if (extracted >= toExtract)
				break
		}

		if (extracted > Fraction.ZERO) {
			copy.count = extracted
			return copy
		}

		return identity.empty()
	}
}

open class PoweredVirtualComponent<T : IStorageStack>(identity: Class<T>, @JvmField val energyProvider: () -> IMatteryEnergyStorage) : VirtualComponent<T>(identity) {
	constructor(identity: Class<T>, energyStorage: IMatteryEnergyStorage) : this(identity, {energyStorage})
	constructor(other: VirtualComponent<T>, energyStorage: IMatteryEnergyStorage) : this(other.storageIdentity(), energyStorage) {
		add(other)
	}

	override fun insertStack(stack: T, simulate: Boolean): T {
		val required = stack.count * identity.energyPerOperation
		val energy = energyProvider()
		val extracted = energy.extractEnergyInner(required, true)

		if (extracted.isZero()) {
			return stack.copy() as T
		}

		if (extracted == required) {
			val leftover = super.insertStack(stack, simulate)

			if (leftover.isEmpty()) {
				if (!simulate) {
					energy.extractEnergyInner(required, false)
				}

				return leftover
			}

			if (!simulate) {
				val requiredNew = (stack.count - leftover.count) * identity.energyPerOperation
				energy.extractEnergyInner(requiredNew, false)
			}

			return leftover
		}

		val stack = stack.copy() as T
		val oldCount = stack.count
		stack.count = extracted / identity.energyPerOperation
		val diff = oldCount - stack.count
		val newRequired = stack.count * identity.energyPerOperation
		val newExtracted = energy.extractEnergyInner(newRequired, true)

		if (newExtracted == newRequired) {
			val leftover = super.insertStack(stack, simulate)

			if (leftover.isEmpty()) {
				if (!simulate) {
					energy.extractEnergyInner(newRequired, false)
				}

				leftover.count = diff
				return leftover
			}

			if (!simulate) {
				val requiredNew = (stack.count - leftover.count) * identity.energyPerOperation
				energy.extractEnergyInner(requiredNew, false)
			}

			leftover.count += diff
			return leftover
		}

		return stack
	}

	override fun extractStack(id: UUID, amount: Fraction, simulate: Boolean): T {
		val required = amount * identity.energyPerOperation
		val energy = energyProvider()
		val extracted = energy.extractEnergyInner(required, true)

		if (extracted.isZero()) {
			return identity.empty
		}

		if (extracted == required) {
			val extractedStack = super.extractStack(id, amount, simulate)

			if (extractedStack.isEmpty()) {
				return extractedStack
			}

			if (!simulate) {
				if (extractedStack.count == amount) {
					energy.extractEnergyInner(required, false)
				} else {
					energy.extractEnergyInner(extractedStack.count * identity.energyPerOperation, false)
				}
			}

			return extractedStack
		}

		val amount = required / identity.energyPerOperation
		val extractedStack = super.extractStack(id, amount, simulate)

		if (extractedStack.isEmpty()) {
			return extractedStack
		}

		if (!simulate) {
			energy.extractEnergyInner(extractedStack.count * identity.energyPerOperation, false)
		}

		return extractedStack
	}
}
